print("import: 'nipy'")
import nipy

print("import: 'nipy.algorithms'")
import nipy.algorithms

print("import: 'nipy.core'")
import nipy.core

print("import: 'nipy.interfaces'")
import nipy.interfaces

print("import: 'nipy.io'")
import nipy.io

print("import: 'nipy.labs'")
import nipy.labs

print("import: 'nipy.modalities'")
import nipy.modalities

